@@ -26,11 +26,11 @@
  * pseudo-terminals and to resize pseudo-terminals.
  */
 
-#include <config.h>
+#include "config.h"
+
+#include <exception>
 
 #include <vte/vte.h>
-#include "vtetypes.hh"
-#include "vtespawn.hh"
 
 #include <errno.h>
 #include <glib.h>
@@ -39,7 +39,12 @@
 
 #include <glib/gi18n-lib.h>
 
+#include "cxx-utils.hh"
+#include "libc-glue.hh"
 #include "pty.hh"
+#include "refptr.hh"
+#include "spawn.hh"
+
 #include "vteptyinternal.hh"
 
 #if !GLIB_CHECK_VERSION(2, 42, 0)
@@ -123,11 +128,10 @@ _vte_pty_get_impl(VtePty* pty)
 /**
  * vte_pty_child_setup:
  * @pty: a #VtePty
- *
- * FIXMEchpe
  */
 void
-vte_pty_child_setup (VtePty *pty)
+vte_pty_child_setup (VtePty *pty) noexcept
+try
 {
         g_return_if_fail(pty != nullptr);
         auto impl = IMPL(pty);
@@ -135,66 +139,9 @@ vte_pty_child_setup (VtePty *pty)
 
         impl->child_setup();
 }
-
-/*
- * __vte_pty_spawn:
- * @pty: a #VtePty
- * @directory: the name of a directory the command should start in, or %NULL
- *   to use the cwd
- * @argv: child's argument vector
- * @envv: a list of environment variables to be added to the environment before
- *   starting the process, or %NULL
- * @spawn_flags: flags from #GSpawnFlags
- * @child_setup: function to run in the child just before exec()
- * @child_setup_data: user data for @child_setup
- * @child_pid: a location to store the child PID, or %NULL
- * @timeout: a timeout value in ms, or %NULL
- * @cancellable: a #GCancellable, or %NULL
- * @error: return location for a #GError, or %NULL
- *
- * Uses g_spawn_async() to spawn the command in @argv. The child's environment will
- * be the parent environment with the variables in @envv set afterwards.
- *
- * Enforces the vte_terminal_watch_child() requirements by adding
- * %G_SPAWN_DO_NOT_REAP_CHILD to @spawn_flags.
- *
- * Note that the %G_SPAWN_LEAVE_DESCRIPTORS_OPEN flag is not supported;
- * it will be cleared!
- *
- * If spawning the command in @working_directory fails because the child
- * is unable to chdir() to it, falls back trying to spawn the command
- * in the parent's working directory.
- *
- * Returns: %TRUE on success, or %FALSE on failure with @error filled in
- */
-gboolean
-_vte_pty_spawn(VtePty *pty,
-               const char *directory,
-               char **argv,
-               char **envv,
-               GSpawnFlags spawn_flags_,
-               GSpawnChildSetupFunc child_setup,
-               gpointer child_setup_data,
-               GPid *child_pid /* out */,
-               int timeout,
-               GCancellable *cancellable,
-               GError **error)
+catch (...)
 {
-        g_return_val_if_fail(VTE_IS_PTY(pty), FALSE);
-
-        auto impl = IMPL(pty);
-        g_return_val_if_fail(impl != nullptr, FALSE);
-
-        return impl->spawn(directory,
-                           argv,
-                           envv,
-                           spawn_flags_,
-                           child_setup,
-                           child_setup_data,
-                           child_pid,
-                           timeout,
-                           cancellable,
-                           error);
+        vte::log_exception();
 }
 
 /**
@@ -215,16 +162,31 @@ gboolean
 vte_pty_set_size(VtePty *pty,
                  int rows,
                  int columns,
-                 GError **error)
+                 GError **error) noexcept
+{
+        /* No way to determine the pixel size; set it to (0, 0), meaning
+         * "undefined".
+         */
+        return _vte_pty_set_size(pty, rows, columns, 0, 0, error);
+}
+
+bool
+_vte_pty_set_size(VtePty *pty,
+                  int rows,
+                  int columns,
+                  int cell_height_px,
+                  int cell_width_px,
+                  GError **error) noexcept
+try
 {
         g_return_val_if_fail(VTE_IS_PTY(pty), FALSE);
         auto impl = IMPL(pty);
         g_return_val_if_fail(impl != nullptr, FALSE);
 
-        if (impl->set_size(rows, columns))
+        if (impl->set_size(rows, columns, cell_height_px, cell_width_px))
                 return true;
 
-        vte::util::restore_errno errsv;
+        auto errsv = vte::libc::ErrnoSaver{};
         g_set_error(error, G_IO_ERROR,
                     g_io_error_from_errno(errsv),
                     "Failed to set window size: %s",
@@ -232,6 +194,11 @@ vte_pty_set_size(VtePty *pty,
 
         return false;
 }
+catch (...)
+{
+        return vte::glib::set_error_from_exception(error);
+}
+
 
 /**
  * vte_pty_get_size:
@@ -250,7 +217,8 @@ gboolean
 vte_pty_get_size(VtePty *pty,
                  int *rows,
                  int *columns,
-                 GError **error)
+                 GError **error) noexcept
+try
 {
         g_return_val_if_fail(VTE_IS_PTY(pty), FALSE);
         auto impl = IMPL(pty);
@@ -259,13 +227,17 @@ vte_pty_get_size(VtePty *pty,
         if (impl->get_size(rows, columns))
                 return true;
 
-        vte::util::restore_errno errsv;
+        auto errsv = vte::libc::ErrnoSaver{};
         g_set_error(error, G_IO_ERROR,
                     g_io_error_from_errno(errsv),
                     "Failed to get window size: %s",
                     g_strerror(errsv));
         return false;
 }
+catch (...)
+{
+        return vte::glib::set_error_from_exception(error);
+}
 
 /**
  * vte_pty_set_utf8:
@@ -282,7 +254,8 @@ vte_pty_get_size(VtePty *pty,
 gboolean
 vte_pty_set_utf8(VtePty *pty,
                  gboolean utf8,
-                 GError **error)
+                 GError **error) noexcept
+try
 {
         g_return_val_if_fail(VTE_IS_PTY(pty), FALSE);
         auto impl = IMPL(pty);
@@ -291,11 +264,15 @@ vte_pty_set_utf8(VtePty *pty,
         if (impl->set_utf8(utf8))
                 return true;
 
-        vte::util::restore_errno errsv;
+        auto errsv = vte::libc::ErrnoSaver{};
         g_set_error(error, G_IO_ERROR, g_io_error_from_errno(errsv),
                     "%s failed: %s", "tc[sg]etattr", g_strerror(errsv));
         return false;
 }
+catch (...)
+{
+        return vte::glib::set_error_from_exception(error);
+}
 
 /**
  * vte_pty_close:
@@ -306,7 +283,7 @@ vte_pty_set_utf8(VtePty *pty,
  * Deprecated: 0.42
  */
 void
-vte_pty_close (VtePty *pty)
+vte_pty_close (VtePty *pty) noexcept
 {
         /* impl->close(); */
 }
@@ -324,17 +301,12 @@ enum {
 static gboolean
 vte_pty_initable_init (GInitable *initable,
                        GCancellable *cancellable,
-                       GError **error)
+                       GError **error) noexcept
+try
 {
         VtePty *pty = VTE_PTY (initable);
         VtePtyPrivate *priv = pty->priv;
 
-        if (cancellable != NULL) {
-                g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                                    "Cancellable initialisation not supported");
-                return FALSE;
-        }
-
         if (priv->foreign_fd != -1) {
                 priv->pty = vte::base::Pty::create_foreign(priv->foreign_fd, priv->flags);
                 priv->foreign_fd = -1;
@@ -343,13 +315,17 @@ vte_pty_initable_init (GInitable *initable,
         }
 
         if (priv->pty == nullptr) {
-                vte::util::restore_errno errsv;
+                auto errsv = vte::libc::ErrnoSaver{};
                 g_set_error(error, G_IO_ERROR, g_io_error_from_errno(errsv),
                             "Failed to open PTY: %s", g_strerror(errsv));
                 return FALSE;
         }
 
-        return TRUE;
+        return !g_cancellable_set_error_if_cancelled(cancellable, error);
+}
+catch (...)
+{
+        return vte::glib::set_error_from_exception(error);
 }
 
 static void
@@ -377,16 +353,20 @@ vte_pty_init (VtePty *pty)
 }
 
 static void
-vte_pty_finalize (GObject *object)
+vte_pty_finalize (GObject *object) noexcept
+try
 {
         VtePty *pty = VTE_PTY (object);
         VtePtyPrivate *priv = pty->priv;
 
-        if (priv->pty != nullptr)
-                priv->pty->unref();
+        auto implptr = vte::base::RefPtr<vte::base::Pty>{priv->pty}; // moved
 
         G_OBJECT_CLASS (vte_pty_parent_class)->finalize (object);
 }
+catch (...)
+{
+        vte::log_exception();
+}
 
 static void
 vte_pty_get_property (GObject    *object,
@@ -484,7 +464,7 @@ vte_pty_class_init (VtePtyClass *klass)
  * Returns: the error domain for VTE PTY errors
  */
 GQuark
-vte_pty_error_quark(void)
+vte_pty_error_quark(void) noexcept
 {
   static GQuark quark = 0;
 
@@ -517,6 +497,10 @@ vte_pty_error_quark(void)
  *
  * Also, you MUST pass the %G_SPAWN_DO_NOT_REAP_CHILD flag.
  *
+ * Note also that %G_SPAWN_STDOUT_TO_DEV_NULL, %G_SPAWN_STDERR_TO_DEV_NULL,
+ * and %G_SPAWN_CHILD_INHERITS_STDIN are not supported, since stdin, stdout
+ * and stderr of the child process will always be connected to the PTY.
+ *
  * Note that you should set the PTY's size using vte_pty_set_size() before
  * spawning the child process, so that the child process has the correct
  * size from the start instead of starting with a default size and then
@@ -528,7 +512,7 @@ vte_pty_error_quark(void)
 VtePty *
 vte_pty_new_sync (VtePtyFlags flags,
                   GCancellable *cancellable,
-                  GError **error)
+                  GError **error) noexcept
 {
         return (VtePty *) g_initable_new (VTE_TYPE_PTY,
                                           cancellable,
@@ -555,7 +539,7 @@ vte_pty_new_sync (VtePtyFlags flags,
 VtePty *
 vte_pty_new_foreign_sync (int fd,
                           GCancellable *cancellable,
-                          GError **error)
+                          GError **error) noexcept
 {
         g_return_val_if_fail(fd != -1, nullptr);
 
@@ -575,91 +559,284 @@ vte_pty_new_foreign_sync (int fd,
  *   its flags changed
  */
 int
-vte_pty_get_fd (VtePty *pty)
+vte_pty_get_fd (VtePty *pty) noexcept
+try
 {
         g_return_val_if_fail(VTE_IS_PTY(pty), FALSE);
-        auto impl = IMPL(pty);
-        g_return_val_if_fail(impl != nullptr, FALSE);
-
-        return impl->fd();
+        return IMPL(pty)->fd();
+}
+catch (...)
+{
+        vte::log_exception();
+        return -1;
 }
 
-typedef struct {
-        VtePty* m_pty;
-        char* m_working_directory;
-        char** m_argv;
-        char** m_envv;
-        GSpawnFlags m_spawn_flags;
-        GSpawnChildSetupFunc m_child_setup;
-        gpointer m_child_setup_data;
-        GDestroyNotify m_child_setup_data_destroy;
-        int m_timeout;
-} AsyncSpawnData;
-
-static AsyncSpawnData*
-async_spawn_data_new (VtePty* pty,
-                      char const* working_directory,
-                      char** argv,
-                      char** envv,
-                      GSpawnFlags spawn_flags,
-                      GSpawnChildSetupFunc child_setup,
-                      gpointer child_setup_data,
-                      GDestroyNotify child_setup_data_destroy,
-                      int timeout)
+static constexpr inline auto
+all_spawn_flags() noexcept
 {
-        auto data = g_new(AsyncSpawnData, 1);
-
-        data->m_pty = (VtePty*)g_object_ref(pty);
-        data->m_working_directory = g_strdup(working_directory);
-        data->m_argv = g_strdupv(argv);
-        data->m_envv = envv ? g_strdupv(envv) : nullptr;
-        data->m_spawn_flags = spawn_flags;
-        data->m_child_setup = child_setup;
-        data->m_child_setup_data = child_setup_data;
-        data->m_child_setup_data_destroy = child_setup_data_destroy;
-        data->m_timeout = timeout;
-
-        return data;
+        return GSpawnFlags(G_SPAWN_LEAVE_DESCRIPTORS_OPEN |
+                           G_SPAWN_DO_NOT_REAP_CHILD |
+                           G_SPAWN_SEARCH_PATH |
+                           G_SPAWN_STDOUT_TO_DEV_NULL |
+                           G_SPAWN_STDERR_TO_DEV_NULL |
+                           G_SPAWN_CHILD_INHERITS_STDIN |
+                           G_SPAWN_FILE_AND_ARGV_ZERO |
+                           G_SPAWN_SEARCH_PATH_FROM_ENVP |
+                           G_SPAWN_CLOEXEC_PIPES |
+                           VTE_SPAWN_NO_PARENT_ENVV |
+                           VTE_SPAWN_NO_SYSTEMD_SCOPE |
+                           VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE);
 }
 
-static void
-async_spawn_data_free(gpointer data_)
+static constexpr inline auto
+forbidden_spawn_flags() noexcept
 {
-        AsyncSpawnData *data = reinterpret_cast<AsyncSpawnData*>(data_);
-
-        g_free(data->m_working_directory);
-        g_strfreev(data->m_argv);
-        g_strfreev(data->m_envv);
-        if (data->m_child_setup_data && data->m_child_setup_data_destroy)
-                data->m_child_setup_data_destroy(data->m_child_setup_data);
-        g_object_unref(data->m_pty);
-
-        g_free(data);
+        return GSpawnFlags(G_SPAWN_LEAVE_DESCRIPTORS_OPEN |
+                           G_SPAWN_STDOUT_TO_DEV_NULL |
+                           G_SPAWN_STDERR_TO_DEV_NULL |
+                           G_SPAWN_CHILD_INHERITS_STDIN);
 }
 
-static void
-async_spawn_run_in_thread(GTask *task,
-                          gpointer object,
-                          gpointer data_,
-                          GCancellable *cancellable)
+static constexpr inline auto
+ignored_spawn_flags() noexcept
 {
-        AsyncSpawnData *data = reinterpret_cast<AsyncSpawnData*>(data_);
+        return GSpawnFlags(G_SPAWN_CLOEXEC_PIPES |
+                           G_SPAWN_DO_NOT_REAP_CHILD);
+}
 
-        GPid pid;
-        GError *error = NULL;
-        if (_vte_pty_spawn(data->m_pty,
-                            data->m_working_directory,
-                            data->m_argv,
-                            data->m_envv,
-                            (GSpawnFlags)data->m_spawn_flags,
-                            data->m_child_setup, data->m_child_setup_data,
-                            &pid,
-                            data->m_timeout,
-                            cancellable,
-                            &error))
-                g_task_return_pointer(task, g_memdup(&pid, sizeof(pid)), g_free);
+static vte::base::SpawnContext
+spawn_context_from_args(VtePty* pty,
+                        char const* working_directory,
+                        char const* const* argv,
+                        char const* const* envv,
+                        int const* fds,
+                        int n_fds,
+                        int const* fd_map_to,
+                        int n_fd_map_to,
+                        GSpawnFlags spawn_flags,
+                        GSpawnChildSetupFunc child_setup,
+                        void* child_setup_data,
+                        GDestroyNotify child_setup_data_destroy)
+{
+        auto context = vte::base::SpawnContext{};
+        context.set_pty(vte::glib::make_ref(pty));
+        context.set_cwd(working_directory);
+        context.set_fallback_cwd(g_get_home_dir());
+        context.set_child_setup(child_setup, child_setup_data, child_setup_data_destroy);
+
+        if ((spawn_flags & G_SPAWN_SEARCH_PATH_FROM_ENVP) ||
+            (spawn_flags & G_SPAWN_SEARCH_PATH))
+                context.set_search_path();
+
+        if (spawn_flags & G_SPAWN_FILE_AND_ARGV_ZERO)
+                context.set_argv(argv[0], argv + 1);
         else
-                g_task_return_error(task, error);
+                context.set_argv(argv[0], argv);
+
+        context.set_environ(envv);
+        if (spawn_flags & VTE_SPAWN_NO_PARENT_ENVV)
+                context.set_no_inherit_environ();
+
+        if (spawn_flags & VTE_SPAWN_NO_SYSTEMD_SCOPE)
+                context.set_no_systemd_scope();
+        if (spawn_flags & VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE)
+                context.set_require_systemd_scope();
+
+        context.add_fds(fds, n_fds);
+        context.add_map_fds(fds, n_fds, fd_map_to, n_fd_map_to);
+
+        return context;
+}
+
+bool
+_vte_pty_spawn_sync(VtePty* pty,
+                    char const* working_directory,
+                    char const* const* argv,
+                    char const* const* envv,
+                    GSpawnFlags spawn_flags,
+                    GSpawnChildSetupFunc child_setup,
+                    gpointer child_setup_data,
+                    GDestroyNotify child_setup_data_destroy,
+                    GPid* child_pid /* out */,
+                    int timeout,
+                    GCancellable* cancellable,
+                    GError** error) noexcept
+try
+{
+        /* These are ignored or need not be passed since the behaviour is the default */
+        g_warn_if_fail((spawn_flags & ignored_spawn_flags()) == 0);
+
+        /* This may be upgraded to a g_return_if_fail in the future */
+        g_warn_if_fail((spawn_flags & forbidden_spawn_flags()) == 0);
+        spawn_flags = GSpawnFlags(spawn_flags & ~forbidden_spawn_flags());
+
+        auto op = vte::base::SpawnOperation{spawn_context_from_args(pty,
+                                                                    working_directory,
+                                                                    argv,
+                                                                    envv,
+                                                                    nullptr, 0,
+                                                                    nullptr, 0,
+                                                                    spawn_flags,
+                                                                    child_setup,
+                                                                    child_setup_data,
+                                                                    child_setup_data_destroy),
+                                            timeout,
+                                            cancellable};
+
+        auto err = vte::glib::Error{};
+        auto rv = op.run_sync(child_pid, err);
+        if (!rv)
+                err.propagate(error);
+
+        return rv;
+}
+catch (...)
+{
+        return vte::glib::set_error_from_exception(error);
+}
+
+/*
+ * _vte_pty_check_envv:
+ * @strv:
+ *
+ * Validates that each element is of the form 'KEY=VALUE'.
+ */
+bool
+_vte_pty_check_envv(char const* const* strv) noexcept
+{
+  if (!strv)
+    return true;
+
+  for (int i = 0; strv[i]; ++i) {
+          const char *str = strv[i];
+          const char *equal = strchr(str, '=');
+          if (equal == nullptr || equal == str)
+                  return false;
+  }
+
+  return true;
+}
+
+/**
+ * vte_pty_spawn_with_fds_async:
+ * @pty: a #VtePty
+ * @working_directory: (allow-none): the name of a directory the command should start
+ *   in, or %NULL to use the current working directory
+ * @argv: (array zero-terminated=1) (element-type filename): child's argument vector
+ * @envv: (allow-none) (array zero-terminated=1) (element-type filename): a list of environment
+ *   variables to be added to the environment before starting the process, or %NULL
+ * @fds: (nullable) (array length=n_fds) (transfer none) (scope call): an array of file descriptors, or %NULL
+ * @n_fds: the number of file descriptors in @fds, or 0 if @fds is %NULL
+ * @map_fds: (nullable) (array length=n_map_fds) (transfer none) (scope call): an array of integers, or %NULL
+ * @n_map_fds: the number of elements in @map_fds, or 0 if @map_fds is %NULL
+ * @spawn_flags: flags from #GSpawnFlags
+ * @child_setup: (allow-none) (scope async): an extra child setup function to run in the child just before exec(), or %NULL
+ * @child_setup_data: (closure child_setup): user data for @child_setup, or %NULL
+ * @child_setup_data_destroy: (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
+ * @timeout: a timeout value in ms, -1 for the default timeout, or G_MAXINT to wait indefinitely
+ * @cancellable: (allow-none): a #GCancellable, or %NULL
+ *
+ * Starts the specified command under the pseudo-terminal @pty.
+ * The @argv and @envv lists should be %NULL-terminated.
+ * The "TERM" environment variable is automatically set to a default value,
+ * but can be overridden from @envv.
+ * @pty_flags controls logging the session to the specified system log files.
+ *
+ * Note also that %G_SPAWN_STDOUT_TO_DEV_NULL, %G_SPAWN_STDERR_TO_DEV_NULL,
+ * and %G_SPAWN_CHILD_INHERITS_STDIN are not supported in @spawn_flags, since
+ * stdin, stdout and stderr of the child process will always be connected to
+ * the PTY. Also %G_SPAWN_LEAVE_DESCRIPTORS_OPEN is not supported; and
+ * %G_SPAWN_DO_NOT_REAP_CHILD will always be added to @spawn_flags.
+ *
+ * If @fds is not %NULL, the child process will map the file descriptors from
+ * @fds according to @map_fds; @n_map_fds must be less or equal to @n_fds.
+ * This function will take ownership of the file descriptors in @fds;
+ * you must not use or close them after this call. All file descriptors in @fds
+ * must have the FD_CLOEXEC flag set on them; it will be unset in the child process
+ * before calling exec.
+ *
+ * Note that all  open file descriptors apart from those mapped as above
+ * will be closed in the child. (If you want to keep some other file descriptor
+ * open for use in the child process, you need to use a child setup function
+ * that unsets the FD_CLOEXEC flag on that file descriptor manually.)
+ *
+ * Beginning with 0.60, and on linux only, and unless %VTE_SPAWN_NO_SYSTEMD_SCOPE is
+ * passed in @spawn_flags, the newly created child process will be moved to its own
+ * systemd user scope; and if %VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE is passed, and creation
+ * of the systemd user scope fails, the whole spawn will fail.
+ * You can override the options used for the systemd user scope by
+ * providing a systemd override file for 'vte-spawn-.scope' unit. See man:systemd.unit(5)
+ * for further information.
+ *
+ * See vte_pty_new(), and vte_terminal_watch_child() for more information.
+ *
+ * Since: 0.62
+ */
+void
+vte_pty_spawn_with_fds_async(VtePty *pty,
+                             char const* working_directory,
+                             char const* const* argv,
+                             char const* const* envv,
+                             int const* fds,
+                             int n_fds,
+                             int const* fd_map_to,
+                             int n_fd_map_to,
+                             GSpawnFlags spawn_flags,
+                             GSpawnChildSetupFunc child_setup,
+                             gpointer child_setup_data,
+                             GDestroyNotify child_setup_data_destroy,
+                             int timeout,
+                             GCancellable *cancellable,
+                             GAsyncReadyCallback callback,
+                             gpointer user_data) noexcept
+try
+{
+        g_return_if_fail(argv != nullptr);
+        g_return_if_fail(argv[0] != nullptr);
+        g_return_if_fail(envv == nullptr || _vte_pty_check_envv(envv));
+        g_return_if_fail(n_fds == 0 || fds != nullptr);
+        for (auto i = int{0}; i < n_fds; ++i)
+                g_return_if_fail(vte::libc::fd_get_cloexec(fds[i]));
+        g_return_if_fail(n_fd_map_to == 0 || fd_map_to != nullptr);
+        g_return_if_fail(n_fds >= n_fd_map_to);
+        g_return_if_fail((spawn_flags & ~all_spawn_flags()) == 0);
+        g_return_if_fail(!child_setup_data || child_setup);
+        g_return_if_fail(!child_setup_data_destroy || child_setup_data);
+        g_return_if_fail(timeout >= -1);
+        g_return_if_fail(cancellable == nullptr || G_IS_CANCELLABLE (cancellable));
+        g_return_if_fail(callback);
+
+        /* These are ignored or need not be passed since the behaviour is the default */
+        g_warn_if_fail((spawn_flags & ignored_spawn_flags()) == 0);
+
+        /* This may be upgraded to a g_return_if_fail in the future */
+        g_warn_if_fail((spawn_flags & forbidden_spawn_flags()) == 0);
+        spawn_flags = GSpawnFlags(spawn_flags & ~forbidden_spawn_flags());
+
+        auto op = new vte::base::SpawnOperation{spawn_context_from_args(pty,
+                                                                        working_directory,
+                                                                        argv,
+                                                                        envv,
+                                                                        fds, n_fds,
+                                                                        fd_map_to, n_fd_map_to,
+                                                                        spawn_flags,
+                                                                        child_setup,
+                                                                        child_setup_data,
+                                                                        child_setup_data_destroy),
+                                                timeout,
+                                                cancellable};
+
+        /* takes ownership of @op */
+        op->run_async((void*)vte_pty_spawn_async, /* tag */
+                      callback,
+                      user_data);
+}
+catch (...)
+{
+        // FIXME: make the function above exception safe. It needs to guarantee
+        // that the callback will be invoked regardless of when the throw occurred.
+
+        vte::log_exception();
 }
 
 /**
@@ -674,31 +851,12 @@ async_spawn_run_in_thread(GTask *task,
  * @child_setup: (allow-none) (scope async): an extra child setup function to run in the child just before exec(), or %NULL
  * @child_setup_data: (closure child_setup): user data for @child_setup, or %NULL
  * @child_setup_data_destroy: (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
- * @timeout: a timeout value in ms, or -1 to wait indefinitely
+ * @timeout: a timeout value in ms, -1 for the default timeout, or G_MAXINT to wait indefinitely
  * @cancellable: (allow-none): a #GCancellable, or %NULL
  *
- * Starts the specified command under the pseudo-terminal @pty.
- * The @argv and @envv lists should be %NULL-terminated.
- * The "TERM" environment variable is automatically set to a default value,
- * but can be overridden from @envv.
- * @pty_flags controls logging the session to the specified system log files.
- *
- * Note that %G_SPAWN_DO_NOT_REAP_CHILD will always be added to @spawn_flags.
- *
- * Note that all open file descriptors will be closed in the child. If you want
- * to keep some file descriptor open for use in the child process, you need to
- * use a child setup function that unsets the FD_CLOEXEC flag on that file
- * descriptor.
- *
- * Beginning with 0.60, and on linux only, and unless %VTE_SPAWN_NO_SYSTEMD_SCOPE is
- * passed in @spawn_flags, the newly created child process will be moved to its own
- * systemd user scope; and if %VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE is passed, and creation
- * of the systemd user scope fails, the whole spawn will fail.
- * You can override the options used for the systemd user scope by
- * providing a systemd override file for 'vte-spawn-.scope' unit. See man:systemd.unit(5)
- * for further information.
- *
- * See vte_pty_new(), g_spawn_async() and vte_terminal_watch_child() for more information.
+ * Like vte_pty_spawn_with_fds_async(), except that this function does not
+ * allow passing file descriptors to the child process. See vte_pty_spawn_with_fds_async()
+ * for more information.
  *
  * Since: 0.48
  */
@@ -714,25 +872,14 @@ vte_pty_spawn_async(VtePty *pty,
                     int timeout,
                     GCancellable *cancellable,
                     GAsyncReadyCallback callback,
-                    gpointer user_data)
+                    gpointer user_data) noexcept
 {
-        g_return_if_fail(argv != nullptr);
-        g_return_if_fail(!child_setup_data || child_setup);
-        g_return_if_fail(!child_setup_data_destroy || child_setup_data);
-        g_return_if_fail(cancellable == nullptr || G_IS_CANCELLABLE (cancellable));
-        g_return_if_fail(callback);
-
-        auto data = async_spawn_data_new(pty,
-                                         working_directory, argv, envv,
-                                         spawn_flags,
-                                         child_setup, child_setup_data, child_setup_data_destroy,
-                                         timeout);
-
-        auto task = g_task_new(pty, cancellable, callback, user_data);
-        g_task_set_source_tag(task, (void*)vte_pty_spawn_async);
-        g_task_set_task_data(task, data, async_spawn_data_free);
-        g_task_run_in_thread(task, async_spawn_run_in_thread);
-        g_object_unref(task);
+        vte_pty_spawn_with_fds_async(pty, working_directory, argv, envv,
+                                     nullptr, 0, nullptr, 0,
+                                     spawn_flags,
+                                     child_setup, child_setup_data, child_setup_data_destroy,
+                                     timeout, cancellable,
+                                     callback, user_data);
 }
 
 /**
@@ -747,27 +894,19 @@ vte_pty_spawn_async(VtePty *pty,
  * Since: 0.48
  */
 gboolean
-vte_pty_spawn_finish(VtePty *pty,
-                     GAsyncResult *result,
-                     GPid *child_pid /* out */,
-                     GError **error)
+vte_pty_spawn_finish(VtePty* pty,
+                     GAsyncResult* result,
+                     GPid* child_pid /* out */,
+                     GError** error) noexcept
 {
-        g_return_val_if_fail (VTE_IS_PTY (pty), FALSE);
-        g_return_val_if_fail (G_IS_TASK (result), FALSE);
-        g_return_val_if_fail(error == nullptr || *error == nullptr, FALSE);
-
-        gpointer pidptr = g_task_propagate_pointer(G_TASK(result), error);
-        if (pidptr == nullptr) {
-                if (child_pid)
-                        *child_pid = -1;
-                return FALSE;
-        }
+        g_return_val_if_fail (VTE_IS_PTY(pty), false);
+        g_return_val_if_fail (G_IS_TASK(result), false);
+        g_return_val_if_fail (g_task_get_source_tag(G_TASK (result)) == vte_pty_spawn_async, false);
+        g_return_val_if_fail(error == nullptr || *error == nullptr, false);
 
+        auto pid = g_task_propagate_int(G_TASK(result), error);
         if (child_pid)
-                *child_pid = *(GPid*)pidptr;
-        if (error)
-                *error = nullptr;
+                *child_pid = pid;
 
-        g_free(pidptr);
-        return TRUE;
+        return pid != -1;
 }
